! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: binimage_mod
!
! Authors:  Brendan Perry - NASA/GSFC (brendan.h.perry@nasa.gov)
!            Comprehensive rewrite and modularization for SAS-19
!           Steven L. Snowden - NASA/GSFC (Original algorithm)
!           Kip Kuntz - NASA/GSFC/JHU (Original algorithm)
!
! Description: Extended Source Analysis Software
!
! Comments:
!
! $Id: binimage_mod.f90,v 1.2 2009/02/20 13:27:08 aibarra Exp $
!-
!  SAS Version 12.0
!  Version 1.1  19 January 2012
!
!  binimage bins images in the XMM-ESAS suite of software.  It operates
!  at the individual image level.
!
!  Input
!      temsc: Threshold for masking
!      idet: Detector selection
!          0: Use the output of comb
!          1: Use a specific exposure
!      prefix: Exposure ID if the detector selection is 1
!      pref: Same as prefix
!      ibin: Binning control
!          1: No binning
!          2: Bin by 2
!          3: Bin by 3
!          4: Bin by 4
!          6: Bin by 6
!          9: Bin by 9
!      iel: Lower energy limit for the energy band in eV
!      ieh: Upper energy limit for the energy band in eV
!      ipart: Particle background control
!          1: include
!          0: do not include
!      iprot: Soft proton control
!          1: include
!          0: do not include
!      iswcx: Solar wind charge exchange control
!          1: include
!          0: do not include
!      imask: Masking control
!          1: include
!          0: do not include
!      maskfile: Mask file name (if absent defaults to using the 
!          exposure map for masking)
! 
!   Examples
!      binimage thresholdmasking=0.02 detector=0 binning=2 elow=350 
!              ehigh=1250 withpartcontrol=1 withsoftcontrol=1
!      binimage thresholdmasking=0.02 detector=1 prefix=1S001 binning=9 
!              elow=350 ehigh=1250 withpartcontrol=1 withsoftcontrol=1
!      binimage thresholdmasking=0.02 detector=0 binning=4 elow=350 
!              ehigh=1250 withpartcontrol=1 withsoftcontrol=0 
!              maskfile=mask.fits
!
!  OUPUT:
!      binned image

module binimage_mod

  use dal
  use param 
  use errorHandling
  use gofutils_mod  ! from goflib, formerly in marq_mod
  use gofmatrix_mod ! from goflib, formerly in marq_mod
  use gofio_mod     ! from goflib, formerly in subsn (SS)

  implicit none 
  private

  character(len=*), parameter :: &
       prefixpar    = "prefix", &
       threshpar    = "maskthresh",&
       combpar      = "withcomb", &
       partpar      = "withpartbkg", &
       sppar        = "withspbkg", &
       swcxpar      = "withswcxbkg", &
       maskpar      = "withmask", &
       maskfilepar  = "maskfile",&
       withbinpar   = "withbinning",&
       binfactpar   = "binfactor",&
       elowpar      = "elow", &
       ehighpar     = "ehigh"

  public binimage

contains

  !---------------------------------------------------------------------
  subroutine binimage
    !---------------------------------------------------------------------
    type(DataSetT) :: inIMG_S, outRATE_S, inMASK_S, inQPB_S, inPROT_S, &
         inSWCX_S, pnOOT_S, inEXP_S, inBKG_S, outSIGMA_S
    type(ArrayT)   :: inIMG_A, outRATE_A, inMASK_A, inQPB_A, inPROT_A, &
         inSWCX_A, pnOOT_A, inEXP_A, inBKG_A, outSIGMA_A
    type(TableT)   :: pnOOT_T, outRATE_T, inPROT_T, inSWCX_T, inBKG_T, &
         inEXP_T, inIMG_T, inMASK_T
    character (len=256) :: inIMG_F, outRATE_F, inMASK_F, inQPB_F, inPROT_F, &
         inSWCX_F, pnOOT_F, inEXP_F, inBKG_F, outSIGMA_F
    character (len=25) :: smode
    character (len=10) :: obsid,  cunitO(2), ctypeO(2), radecsysO, cunit(2), &
         ctype(2), cunitE(2), ctypeE(2), radecsysE, radecsys, style
    character (len=7)  :: name
    character (len=6)  :: filter
    character (len=5)  :: prefix, expid, celow, cehigh, exptype, suffix
    character (len=2)  :: inst
    integer :: xSize=900, ySize=900, xSizeD, ySizeD
    integer :: i, j, k, ii, jj, i1, i2, j1, j2, jj1, jj2
    integer :: status, ios, ibin, ir, irr, ioff, outlun, xdim, ydim, dims(2)
    integer :: type, frmt, revl, nR, el, eh, elow, ehigh
    integer :: binfact, typeA
    integer, allocatable, dimension (:) :: n
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: aDI32, inIMG_JP, pnOOT_JP
    !
    logical :: oot, warndie, withpart, withprot, withswcx, withmask, withbin, &
         withoot, withcomb
    !
    real :: backO, areaO, expoO, backD, areaD, expoD
    real :: imcounts, qpbcounts, expcounts, protcounts, swcxcounts
    real :: maskthresh
    real :: xcen, ycen ! why reals???!!!
    real :: b, blast, c, clast, e, elast, ex, num, numlast, ootsca, r, &
         rad, sca, sn, temp, t, tlim, trad, xmax, ymax, s
    real(kind=double) :: crvalO(2), cdeltO(2), crpixO(2), crotaO(2), &
         ra_pntO, dec_pntO, pa_pntO, crval(2), cdelt(2), crpix(2), crota(2), &
         ra_pnt, dec_pnt, pa_pnt, crvalE(2), cdeltE(2), crpixE(2), crotaE(2), &
         ra_pntE, dec_pntE, pa_pntE
    real, dimension(:), allocatable :: bck, cts, expos, x, y, y2, ys ! why 50???!!!
    real, dimension(:), allocatable :: xh, yh ! why 10000/1000 in orig code???!!!
    real, dimension(:,:), allocatable :: mask, expose, back, counts, ootcounts, &
         prot, swcx, pnOOT, rate, sigma
    real(kind=single), dimension(:,:), pointer :: aDR32, inIMG_RP, inMASK_RP, &
         inQPB_RP, inPROT_RP, inSWCX_RP, inEXP_RP, outRATE_RP, outSIGMA_RP, &
         inBKG_RP, inOOT_RP
    real(kind=double), dimension(:,:), pointer :: aDR64
    !
    withoot = .FALSE. ! for now
    oot     = .FALSE. ! for now
    !
    ! Get the input parameters
    !
    prefix = stringParameter(prefixpar)
    withcomb    = booleanParameter(combpar)
    elow        = intParameter(elowpar)         ! IEL
    ehigh       = intParameter(ehighpar)        ! IEH
    write(celow,'(I5)') elow   ! character form
    write(cehigh,'(I5)') ehigh ! character form
    withpart    = booleanParameter(partpar)     ! IPART
    withprot    = booleanParameter(sppar)       ! IPROT
    withswcx    = booleanParameter(swcxpar)     ! ISWCX
    withmask    = .FALSE.
    withmask    = booleanParameter(maskpar)
    if (withmask) inMASK_F   = stringParameter(maskfilepar) ! OLDNEW: MASKFILE
    binfact     = 1
    withbin     = booleanParameter(withbinpar)
    if (withbin) binfact = intParameter(binfactpar)     ! OLDNEW: ibin
    !
    !  SS: Enter the fractional limit for template, the program will exclude
    !  all pixels in the field where the mask value is less than the 
    !  fractional limit times the average non-zero value of the mask
    !
    maskthresh  = realParameter(threshpar)                  ! OLDNEW: temsc
    !
    ! create the filenames, fill the arrays
    ! call check_clobber ???!!!
    ! OUTPUT FILES:
    ! OLDNEW: adapt-350-1100.fits/mos1S001-adaptimsky-350-1100.fits
    ! OLDNEW: size-350-1100.fits/mos1S001-sizeimsky-350-1100.fits
    suffix = ".fits"
    style = "rateim" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,outRATE_F) 
    style = "sigim" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,outSIGMA_F)
    !
    ! INPUT FILES:
    ! Get the FOV image and keywords
    ! OLDNEW: comb-obj-im-350-1100.fits/comb-fovim-350-1100.fits use sky???!!! or...
    ! OLDNEW: mos1S001-obj-im-350-1100.fits/mos1S001-fovimsky-350-1100.fits
    style = "fovim" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inIMG_F) ! input IMAGE
    write(errstr,*)'***** Reading file: ',trim(adjustl(inIMG_F))
    call message(AppMsg,Verbose)
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    call check_open_set(inIMG_F,inIMG_S,1,status,warndie) ! goflib/gofio_mod
    type = 3 ! image
    ! extract image array and coordinate keys (these are in gofio/goflib also)
    !
    ! BHP: ordinary type FOVIM is Int32, otherwise all other IMGs used in adapt are
    !      Real32
    ! fill the counts array:
    allocate(counts(xSize,ySize))
    counts = 0.
    if (withcomb) then ! comb type files are R*4, non-comb files are I*4
       call read_2Darray(inIMG_S,inIMG_A,inIMG_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             counts(i,j) = inIMG_RP(i-1,j-1)
          enddo
       enddo
    else
       call read_2Darray(inIMG_S,inIMG_A,aDR32,aDR64,aDI16,inIMG_JP,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             counts(i,j) = real(inIMG_JP(i-1,j-1))
          enddo
       enddo
    endif
    call coord_keys (inIMG_S, ra_pntO, dec_pntO, pa_pntO, crvalO, cdeltO, &
         cunitO, ctypeO, crpixO, crotaO, radecsysO, warndie)
    !
    ! BHP: get the PN OOT, apply to counts array (if applicable)
    ! comb-oot???!!!
    ! OLDNEW: pnS003-obj-im-oot-350-1100.fits/pnS003-fovimootsky-350-1100.fits
    if (trim(adjustl(inst)) .eq. 'PN') then ! always???!!! or param withoot???!!!
       write(errstr,*)'***** Reading file: ',trim(adjustl(pnOOT_F))
       call message(AppMsg,Verbose)
       oot = .TRUE.
       suffix = ".fits"
       style = "fovim" 
       call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,pnOOT_F) ! input IMAGE
       write(errstr,*)'PN OOT file = ',trim(adjustl(pnOOT_F))
       call message(AppMsg,Verbose)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(pnOOT_F,pnOOT_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,pnOOT_S,pnOOT_T,pnOOT_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(pnOOT_S,pnOOT_A,aDR32,aDR64,aDI16,pnOOT_JP,xSizeD,ySizeD,typeA)
       ! SS: This is a bit painful, but we need to know the SUBMODE
       !     keyword for the proper scaling of the OOT events
       ! BHP: Not anymore. Just need the smode from the header, then call
       !      getootsca from goflib...
       call getootsca(smode,ootsca,warndie) ! goflib/gofutils OOT scale factor
       ! apply oot to counts array (no need to put into own array, never gets used again!)
       do i=1,xSize
          do j=1,ySize
             counts(i,ii) = counts(i,ii) - ootsca * real(pnOOT_JP(i-1,ii-1))
          enddo
       enddo
       call release(pnOOT_S)
       withoot = .TRUE.
    endif
    ! Get the exposure map and keywords
    ! OLDNEW: comb-exp-im-350-1100.fits/comb-fovim-350-1100.fits sky???!!!
    ! OLDNEW: mos1S001-exp-im-350-1100.fits/mos1S001-expimsky-350-1100.fits
    style = "expim" 
    call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inEXP_F) ! input exposure map
    allocate(expose(xSize,ySize))
    expose = 0.
    write(errstr,*)'***** Reading file: ',trim(adjustl(inEXP_F))
    call message(AppMsg,Verbose)
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    call check_open_set(inEXP_F,inEXP_S,1,status,warndie) ! goflib/gofio_mod
    type = 3 ! image
    ! extract image array and coordinate keys (these are in gofio/goflib also)
    call read_header_info(type,inEXP_S,inEXP_T,inEXP_A,inst,obsid,expid,&
         smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
    call read_2Darray(inEXP_S,inEXP_A,inEXP_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
    call coord_keys (inIMG_S, ra_pntE, dec_pntE, pa_pntE, crvalE, cdeltE, &
         cunitE, ctypeE, crpixE, crotaE, radecsysE, warndie)
    do i=1,xSize
       do j=1,ySize
          expose(i,j) = inEXP_RP(i-1,j-1)
       enddo
    enddo
    call release(inEXP_S)
    ! Optional images...
    ! Get the mask and keywords (if selected)
    ! OLDNEW: comb-mask-im-350-1100.fits/mos1S001-maskimsky-350-1100.fits
    ! OLDNEW: mos1S001-mask-im-350-1100.fits/mos1S001-maskimsky-350-1100.fits
    allocate(mask(xSize,ySize))
    mask = 0.
    if (withmask) then 
       style = "maskim" 
       call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inMASK_F) ! input mask
       write(errstr,*)'***** Reading file: ',trim(adjustl(inMASK_F))
       call message(AppMsg,Verbose)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(inMASK_F,inMASK_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,inMASK_S,inMASK_T,inMASK_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(inMASK_S,inMASK_A,inMASK_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             mask(i,j) = inMASK_RP(i-1,j-1)
          enddo
       enddo
       call release(inMASK_S)
    else
       do i=1,xSize
          do j=1,ySize
             mask(i,j) = expose(i,j)
          enddo
       enddo
    endif
    ! Get particle background backim image and keywords (if selected)
    ! OLDNEW: comb-back-im-350-1100.fits/comb-bkgim-350-1100.fits sky???!!!
    ! OLDNEW: mos1S001-back-im-sky-350-1100.fits/mos1S001-bkgimsky-350-1100.fits
    allocate(back(xSize,ySize))
    back = 0.
    if (withpart) then 
       style = "bkgim" 
       call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inBKG_F) ! input background map
       write(errstr,*)'***** Reading file: ',trim(adjustl(inBKG_F))
       call message(AppMsg,Verbose)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(inBKG_F,inBKG_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,inBKG_S,inBKG_T,inBKG_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(inBKG_S,inBKG_A,inBKG_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             back(i,j) = inBKG_RP(i-1,j-1)
          enddo
       enddo
       call release(inBKG_S)
    endif
    ! Get proton-produced soft proton image and keywords
    ! OLDNEW: comb-prot-im-350-1100.fits/comb-protim-350-1100.fits sky???!!!    
    ! OLDNEW: mos1S001-prot-im-sky-350-1100.fits/mos1S001-protimsky-350-1100.fits
    allocate(prot(xSize,ySize))
    prot = 0.
    if (withprot) then 
       style = "protim" 
       call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inPROT_F) ! input SP map
       write(errstr,*)'***** Reading file: ',trim(adjustl(inPROT_F))
       call message(AppMsg,Verbose)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(inPROT_F,inPROT_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,inPROT_S,inPROT_T,inPROT_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(inPROT_S,inPROT_A,inPROT_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             prot(i,j) = inPROT_RP(i-1,j-1)
          enddo
       enddo
       call release(inPROT_S)
    endif
    ! Get swcx-produced image and keywords
    ! OLDNEW: comb-swcx-im-350-1100.fits/comb-swcxim-350-1100.fits sky???!!!    
    ! OLDNEW: mos1S001-swcx-im-sky-350-1100.fits/mos1S001-swcximsky-350-1100.fits
    allocate(swcx(xSize,ySize))
    swcx = 0.
    if (withswcx) then 
       style = "swcxim" 
       call makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,inSWCX_F) ! input SWCX map
       write(errstr,*)'***** Reading file: ',trim(adjustl(inSWCX_F))
       call message(AppMsg,Verbose)
       status = 0
       warndie = .TRUE. ! croak if not found or unreadable
       call check_open_set(inSWCX_F,inSWCX_S,1,status,warndie) ! goflib/gofio_mod
       type = 3 ! image
       ! extract image array and coordinate keys (these are in gofio/goflib also)
       call read_header_info(type,inSWCX_S,inSWCX_T,inSWCX_A,inst,obsid,expid,&
            smode,frmt,filter,revl,backD,areaD,expoD,el,eh,nR,warndie)
       call read_2Darray(inSWCX_S,inSWCX_A,inSWCX_RP,aDR64,aDI16,aDI32,xSizeD,ySizeD,typeA)
       do i=1,xSize
          do j=1,ySize
             swcx(i,j) = inSWCX_RP(i-1,j-1)
          enddo
       enddo
       call release(inSWCX_S)
    endif
    !
    call printParams(inIMG_F,inEXP_F,inBKG_F,outRATE_F,outSIGMA_F,withcomb,celow,cehigh,&
         withpart,withprot,withswcx,withmask,inMASK_F,withbin,binfact,maskthresh)
    !
    !  SS: Find out how many counts we have to play with and the exposure center
    ! 
    imcounts   = 0.0 ! RNC
    qpbcounts  = 0.0 ! RNB
    protcounts = 0.0 ! RNP
    swcxcounts = 0.0 ! RNS
    expcounts  = 0.0 ! REX
    ex = 0.0
    do i=1,xSize
       do ii=1,ySize
          imcounts = imcounts + counts(i,ii)
          if (withpart) qpbcounts = qpbcounts + back(i,ii)
          if (withprot) protcounts = protcounts + prot(i,ii)
          if (withswcx) swcxcounts = swcxcounts + swcx(i,ii)
          expcounts = expcounts + expose(i,ii)
          if (withprot .and. withswcx) then
             back(i,ii) = back(i,ii) + prot(i,ii) + swcx(i,ii)
          endif
          if(expose(i,ii) .gt. ex) then
             ex = expose(i,ii)
             xcen = i
             ycen = ii
          endif
       enddo
    enddo
    ! BHP: these are exactly the same, old v new
    write(errstr,*) "Total Counts   : ", imcounts 
    call message(AppMsg,Always)
    write(errstr,*) "Exposure       : ", expcounts
    call message(AppMsg,Always)
    write(errstr,*) "QPB Counts     : ", qpbcounts
    call message(AppMsg,Always)
    write(errstr,*) "Proton Counts  : ", protcounts
    call message(AppMsg,Always)
    write(errstr,*) "SWCX Counts    : ", swcxcounts
    call message(AppMsg,Always)
    !
    ioff = mod(xSize,binfact)/2 ! 900
    xdim = xSize/binfact
    ydim = ySize/binfact ! OLDNEW: ibin
    sca = -1.0 / (cdeltO(1)*cdeltO(2)) ! or cdeltE???!!!
    ! print*,ioff,xdim,ydim,sca
    !
    ! SS: Read in the template
    ! SS: Use the exposure maps as the template
    ! OLD: IF(MASKFILE .EQ. '      ') THEN
    !
    ! SS: Set the template limit
    !
    j = 0
    tlim = 0.
    do i=1,xSize
       do ii=1,ySize
          if(mask(i,ii) .gt. 0) then
             tlim = tlim + mask(i,ii)
             j = j + 1
          endif
       enddo
    enddo
    !    print*,'tlim, maskthresh, j',tlim,maskthresh,j,binfact,xSize
    tlim = maskthresh * tlim / j ! OLDNEW: TEMSC*TLIM/J
    !
    !
    !  SS: Set the binning limits
    !
    i2 = xSize / binfact ! OLDNEW:    I2 = 900/IBIN
    ioff = mod(900,binfact)/2 ! was iex, but same as adapt's ioff!
    allocate(rate(i2,i2),sigma(i2,i2))
    rate = 0.
    sigma = 0.
    b = 0.
    c = 0.
    e = 0.
    do i=1,i2
       j2 = i*binfact + ioff
       j1 = j2 - binfact + 1
       do ii=1,i2
          jj2 = ii*binfact + ioff
          jj1 = jj2 - binfact + 1
          b = 0.
          c = 0.
          e = 0.
          do j=j1,j2
             do jj=jj1,jj2
                if(mask(j,jj) .gt. tlim) then
                   b = b + back(j,jj)
                   c = c + counts(j,jj)
                   e = e + expose(j,jj)
                endif
             enddo
          enddo
          if(e .gt. 0.0) then
             r = sca*(c - b)/e
             if(c .gt. 0.0) then
                s = sca*sqrt(c)/e
             else
                s = 0.0
             endif
          else
             r = 0.0
             s = 0.0
          endif
          rate(i,ii) = r
          sigma(i,ii) = s
          if (rate(i,ii) .ne. 0) then 
             write(52,*)i,ii,rate(i,ii)
          endif
          if (sigma(i,ii) .ne. 0) then 
             write(53,*)i,ii,sigma(i,ii)
          endif
       enddo
    enddo
    !
    ! Open the new file rate.fits and sigma.fits, write out
    !
    cdelt(1) = binfact * cdeltE(1) ! fix this???!!!
    cdelt(2) = binfact * cdeltE(2)
    crpix(1) = xdim / 2 + 0.5
    crpix(2) = ydim / 2 + 0.5
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    call check_clobber(outRATE_F,warndie)
    ! 3 = CREATE
    call check_open_set(outRATE_F,outRATE_S,3,status,warndie) ! goflib/gofio_mod
    type = 3 ! 3 = image
    name = "PRIMARY"
    outRATE_A = addArray(outRATE_S, trim(name), REAL32, (/xdim,ydim/))
    outRATE_RP => real32Array2Data(outRATE_A)    
    !
    status = 0
    warndie = .TRUE. ! croak if not found or unreadable
    call check_clobber(outSIGMA_F,warndie)
    ! 3 = CREATE
    call check_open_set(outSIGMA_F,outSIGMA_S,3,status,warndie) ! goflib/gofio_mod
    type = 3 ! 3 = image
    name = "PRIMARY"
    outSIGMA_A = addArray(outSIGMA_S, trim(name), REAL32, (/xdim,ydim/))
    outSIGMA_RP => real32Array2Data(outSIGMA_A)
    !
    ! or copyAttributes then change the 'new ones'???!!!
    !
    call setAttribute(outRATE_A, 'CONTENT', "IMAGE")
    call setAttribute(outRATE_A, 'TELESCOP', "XMM", "mission name")
    call setAttribute(outRATE_A, 'INSTRUME', "EPIC","instrument name")
    call setAttribute(outRATE_A, 'OBS_MODE', "POINTING","obs mode: POINTING, SLEW, OR SCAN")
    call setAttribute(outRATE_A, 'RADECSYS', radecsysE,"Equitorial system reference")
    call setAttribute(outRATE_A, 'EQUINOX', 2000, "Equinox")
    !call setAttribute(outRATE_A, 'BZERO', 0., "") how to write???!!!
    !call setAttribute(outRATE_A, 'BSCALE', 1., "")
    !call setAttribute(outRATE_A, 'BUNIT', "c/s/deg2", "Units of data")
    call setAttribute(outRATE_A, 'CRPIX1', crpix(1), "Reference pixel")
    call setAttribute(outRATE_A, 'CTYPE1', ctypeE(1), "Projection")
    call setAttribute(outRATE_A, 'CRVAL1', crvalE(1), "Central Longitude")
    call setAttribute(outRATE_A, 'CDELT1', cdelt(1), "WCS pixel size")
    call setAttribute(outRATE_A, 'CUNIT1', "deg", "Units of coordinate")
    call setAttribute(outRATE_A, 'CRPIX2', crpix(2), "Reference pixel")
    call setAttribute(outRATE_A, 'CTYPE2', ctypeE(2), "Projection")
    call setAttribute(outRATE_A, 'CRVAL2', crvalE(2), "Central Latitude")    
    call setAttribute(outRATE_A, 'CDELT2', cdelt(2), "WCS pixel size")
    call setAttribute(outRATE_A, 'CUNIT2', "deg", "Units of coordinate")

    call setAttribute(outSIGMA_A, 'CONTENT', "IMAGE")
    call setAttribute(outSIGMA_A, 'TELESCOP', "XMM", "mission name")
    call setAttribute(outSIGMA_A, 'INSTRUME', "EPIC","instrument name")
    call setAttribute(outSIGMA_A, 'OBS_MODE', "POINTING","obs mode: POINTING, SLEW, OR SCAN")
    call setAttribute(outSIGMA_A, 'RADECSYS', radecsysE,"Equitorial system reference")
    call setAttribute(outSIGMA_A, 'EQUINOX', 2000, "Equinox")
    !call setAttribute(outSIGMA_A, 'BZERO', 0., "") how to write???!!!
    !call setAttribute(outSIGMA_A, 'BSCALE', 1., "")
    !call setAttribute(outSIGMA_A, 'BUNIT', "c/s/deg2", "Units of data")
    call setAttribute(outSIGMA_A, 'CRPIX1', crpix(1), "Reference pixel")
    call setAttribute(outSIGMA_A, 'CTYPE1', ctypeE(1), "Projection")
    call setAttribute(outSIGMA_A, 'CRVAL1', crvalE(1), "Central Longitude")
    call setAttribute(outSIGMA_A, 'CDELT1', cdelt(1), "WCS pixel size")
    call setAttribute(outSIGMA_A, 'CUNIT1', "deg", "Units of coordinate")
    call setAttribute(outSIGMA_A, 'CRPIX2', crpix(2), "Reference pixel")
    call setAttribute(outSIGMA_A, 'CTYPE2', ctypeE(2), "Projection")
    call setAttribute(outSIGMA_A, 'CRVAL2', crvalE(2), "Central Latitude")    
    call setAttribute(outSIGMA_A, 'CDELT2', cdelt(2), "WCS pixel size")
    call setAttribute(outSIGMA_A, 'CUNIT2', "deg", "Units of coordinate")
    ! fill the final array pointers opened above
    do i=1,xdim
       do j=1,ydim
          outRATE_RP(i-1,j-1)  = rate(i,j)
          outSIGMA_RP(i-1,j-1) = sigma(i,j)
          !print*,i,j,outRATE_RP(i-1,j-1)
       enddo
    enddo
    ! close and seal the output files
    call release(outRATE_S)
    call release(outSIGMA_S)

    return

  end subroutine binimage
  !
  ! Subroutine section
  !
  subroutine makeFileNames(prefix,celow,cehigh,withcomb,style,suffix,oot,file)
    !
    character(len=*), intent(in) :: prefix, suffix, style, celow, cehigh
    character(len=256), intent(out) :: file
    character(len=3) :: pre, sky
    !
    logical :: withcomb, oot
    !
    sky = 'sky'
    if (suffix .eq. '.qdp') sky = ''
    !
    if (withcomb) then ! should infile names have 'sky' in these???!!!
       file = "comb-" // trim(adjustl(style)) //'-' // trim(adjustl(celow)) &
            // '-' // trim(adjustl(cehigh)) // trim(adjustl(suffix))
    else       
       pre = 'mos'
       if ((prefix(1:1) .ne. '1') .and. (prefix(1:1) .ne. '2')) then
          pre = 'pn'
       endif
       if (oot) then 
          file = trim(adjustl(pre)) // trim(adjustl(prefix)) // &
               '-' // trim(adjustl(style)) // "ootsky" // '-' // &
               trim(adjustl(celow)) // '-' // trim(adjustl(cehigh)) &
               // trim(adjustl(suffix))
       else
          file = trim(adjustl(pre)) // trim(adjustl(prefix)) // &
               '-' // trim(adjustl(style)) // trim(adjustl(sky)) // '-' // &
               trim(adjustl(celow)) // '-' // trim(adjustl(cehigh)) &
               // trim(adjustl(suffix))
       endif
    endif

    return

  end subroutine makeFileNames
  !
  !
  !
  subroutine printParams(inIMG_F,inEXP_F,inBKG_F,outRATE_F,outSIGMA_F,withcomb,celow,cehigh,&
       withpart,withprot,withswcx,withmask,inMASK_F,withbin,binfact,maskthresh)
    !
    character(len=*) :: inIMG_F, inEXP_F, inBKG_F, outRATE_F, outSIGMA_F, inMASK_F, celow, cehigh
    character(len=4) :: cbin
    !
    integer :: binfact
    !
    logical, intent(in) :: withcomb, withpart, withprot, withswcx, withmask, &
         withbin
    !
    real :: smoofact, maskthresh
    write(cbin,'(I4)') binfact   ! character form

    write(errstr,*) "Input parameters:"
    call message(AppMsg,Always)
    write(errstr,*) "Input IMAGE filename             : ",trim(inIMG_F)    ! OLDNEW: INFILE
    call message(AppMsg,Always)
    if(withcomb) then
       write(errstr,*) "Use combimage-produced files?    : ",withcomb 
       call message(AppMsg,Always)
    endif
    write(errstr,*) "Input EXP IMAGE filename         : ",trim(inEXP_F)    ! OLDNEW: INFILE!
    call message(AppMsg,Always)
    write(errstr,*) "Input BKG IMAGE filename         : ",trim(inBKG_F)    ! OLDNEW: INFILE!
    call message(AppMsg,Always)
    write(errstr,*) "Output RATE IMAGE filename       : ",trim(outRATE_F)   ! OLDNEW: INFILE!
    call message(AppMsg,Always)
    write(errstr,*) "Output SIGMA IMAGE filename      : ",trim(outSIGMA_F)   ! OLDNEW: INFILE!
    call message(AppMsg,Always)
    write(errstr,*) "Apply source removal mask?       : ",withmask
    call message(AppMsg,Always)
    if (withmask) then
       write(errstr,*) "Source removal mask              : ",trim(adjustl(inMASK_F))
       call message(AppMsg,Always)
    endif
    write(errstr,*) "Selected energy lower limit      : ",trim(adjustr(celow))
    call message(AppMsg,Always)
    write(errstr,*) "Selected energy upper limit      : ",trim(adjustr(cehigh))
    call message(AppMsg,Always)
    write(errstr,*) "Apply particle (QPB) count data? : ",withpart
    call message(AppMsg,Always)
    write(errstr,*) "Apply soft proton count data?    : ",withprot
    call message(AppMsg,Always)
    write(errstr,*) "Apply SWCX count data?           : ",withswcx
    call message(AppMsg,Always)
    write(errstr,*) "Bin the data?                    : ",withbin
    call message(AppMsg,Always)
    write(errstr,*) "Bin factor                       :  ",trim(adjustl(cbin))
    call message(AppMsg,Always)
    write(errstr,*) "Fractional limit for template    : ", maskthresh
    call message(AppMsg,Always)

    return

  end subroutine printparams

end module binimage_mod
